"""
pytransport - Royal Holloway utility to manipulate TRANSPORT data and models.

Authors:

 * William Shields
 * Jochem Snuverink

Copyright Royal Holloway, University of London 2018.

"""

__version__ = "1.1"

import _General
import Compare
import Convert
import Data
import Reader

__all__ = ['Compare',
           'Convert',
           'Data',
           'Reader']
